@@ -10,7 +10,6 @@
 #include "radv_sdma.h"
 #include "util/macros.h"
 #include "util/u_memory.h"
-#include "radv_buffer.h"
 #include "radv_cs.h"
 #include "radv_formats.h"
 
@@ -165,10 +164,9 @@ radv_sdma_get_bpe(const struct radv_image *const image, VkImageAspectFlags aspec
 }
 
 struct radv_sdma_surf
-radv_sdma_get_buf_surf(const struct radv_buffer *const buffer, const struct radv_image *const image,
-                       const VkBufferImageCopy2 *const region, const VkImageAspectFlags aspect_mask)
+radv_sdma_get_buf_surf(uint64_t buffer_va, const struct radv_image *const image, const VkBufferImageCopy2 *const region)
 {
-   assert(util_bitcount(aspect_mask) == 1);
+   assert(util_bitcount(region->imageSubresource.aspectMask) == 1);
 
    const unsigned pitch = (region->bufferRowLength ? region->bufferRowLength : region->imageExtent.width);
    const unsigned slice_pitch =
@@ -179,7 +177,7 @@ radv_sdma_get_buf_surf(const struct radv_buffer *const buffer, const struct radv
    const uint32_t bpe = radv_sdma_get_bpe(image, region->imageSubresource.aspectMask);
 
    const struct radv_sdma_surf info = {
-      .va = radv_buffer_get_va(buffer->bo) + buffer->offset + region->bufferOffset,
+      .va = buffer_va + region->bufferOffset,
       .pitch = pitch,
       .slice_pitch = slice_pitch,
       .bpp = bpe,
@@ -193,29 +191,28 @@ radv_sdma_get_buf_surf(const struct radv_buffer *const buffer, const struct radv
 
 static uint32_t
 radv_sdma_get_metadata_config(const struct radv_device *const device, const struct radv_image *const image,
-                              const struct radeon_surf *const surf, const VkImageSubresourceLayers subresource,
-                              const VkImageAspectFlags aspect_mask)
+                              const struct radeon_surf *const surf, const VkImageSubresourceLayers subresource)
 {
    const struct radv_physical_device *pdev = radv_device_physical(device);
 
-   if (!pdev->info.sdma_supports_compression ||
-       !(radv_dcc_enabled(image, subresource.mipLevel) || radv_image_has_htile(image))) {
-      return 0;
-   }
-
-   const VkFormat format = vk_format_get_aspect_format(image->vk.format, aspect_mask);
-   const struct util_format_description *desc = vk_format_description(format);
+   const VkFormat format = vk_format_get_aspect_format(image->vk.format, subresource.aspectMask);
 
    const uint32_t data_format = ac_get_cb_format(pdev->info.gfx_level, radv_format_to_pipe_format(format));
    const uint32_t alpha_is_on_msb = ac_alpha_is_on_msb(&pdev->info, radv_format_to_pipe_format(format));
-   const uint32_t number_type = radv_translate_buffer_numformat(desc, vk_format_get_first_non_void_channel(format));
-   const uint32_t surface_type = radv_sdma_surface_type_from_aspect_mask(aspect_mask);
+   const uint32_t number_type = ac_get_cb_number_type(radv_format_to_pipe_format(format));
+   const uint32_t surface_type = radv_sdma_surface_type_from_aspect_mask(subresource.aspectMask);
    const uint32_t max_comp_block_size = surf->u.gfx9.color.dcc.max_compressed_block_size;
-   const uint32_t max_uncomp_block_size = radv_get_dcc_max_uncompressed_block_size(device, image);
-   const uint32_t pipe_aligned = surf->u.gfx9.color.dcc.pipe_aligned;
+   const uint32_t pipe_aligned = radv_htile_enabled(image, subresource.mipLevel) || surf->u.gfx9.color.dcc.pipe_aligned;
 
-   return data_format | alpha_is_on_msb << 8 | number_type << 9 | surface_type << 12 | max_comp_block_size << 24 |
-          max_uncomp_block_size << 26 | pipe_aligned << 31;
+   if (pdev->info.sdma_ip_version >= SDMA_7_0) {
+      return SDMA7_DCC_DATA_FORMAT(data_format) | SDMA7_DCC_NUM_TYPE(number_type) | SDMA7_DCC_READ_CM(2) |
+             SDMA7_DCC_MAX_COM(max_comp_block_size) | SDMA7_DCC_MAX_UCOM(1);
+   } else {
+      return SDMA5_DCC_DATA_FORMAT(data_format) | SDMA5_DCC_ALPHA_IS_ON_MSB(alpha_is_on_msb) |
+             SDMA5_DCC_NUM_TYPE(number_type) | SDMA5_DCC_SURF_TYPE(surface_type) |
+             SDMA5_DCC_MAX_COM(max_comp_block_size) | SDMA5_DCC_MAX_UCOM(V_028C78_MAX_BLOCK_SIZE_256B) |
+             SDMA5_DCC_PIPE_ALIGNED(pipe_aligned);
+   }
 }
 
 static uint32_t
@@ -263,17 +260,17 @@ radv_sdma_get_tiled_header_dword(const struct radv_device *const device, const s
 
 struct radv_sdma_surf
 radv_sdma_get_surf(const struct radv_device *const device, const struct radv_image *const image,
-                   const VkImageSubresourceLayers subresource, const VkOffset3D offset,
-                   const VkImageAspectFlags aspect_mask)
+                   const VkImageSubresourceLayers subresource, const VkOffset3D offset)
 {
-   assert(util_bitcount(aspect_mask) == 1);
+   assert(util_bitcount(subresource.aspectMask) == 1);
 
    const struct radv_physical_device *pdev = radv_device_physical(device);
-   const unsigned plane_idx = radv_plane_from_aspect(aspect_mask);
+   const unsigned plane_idx = radv_plane_from_aspect(subresource.aspectMask);
    const unsigned binding_idx = image->disjoint ? plane_idx : 0;
    const struct radeon_surf *const surf = &image->planes[plane_idx].surface;
-   const uint64_t va = radv_image_get_va(image, binding_idx);
-   const uint32_t bpe = radv_sdma_get_bpe(image, aspect_mask);
+   const struct radv_image_binding *binding = &image->bindings[binding_idx];
+   const uint64_t va = binding->addr;
+   const uint32_t bpe = radv_sdma_get_bpe(image, subresource.aspectMask);
    struct radv_sdma_surf info = {
       .extent =
          {
@@ -296,8 +293,8 @@ radv_sdma_get_surf(const struct radv_device *const device, const struct radv_ima
       .is_3d = surf->u.gfx9.resource_type == RADEON_RESOURCE_3D,
    };
 
-   const uint64_t surf_offset =
-      (aspect_mask == VK_IMAGE_ASPECT_STENCIL_BIT) ? surf->u.gfx9.zs.stencil_offset : surf->u.gfx9.surf_offset;
+   const uint64_t surf_offset = (subresource.aspectMask == VK_IMAGE_ASPECT_STENCIL_BIT) ? surf->u.gfx9.zs.stencil_offset
+                                                                                        : surf->u.gfx9.surf_offset;
 
    if (surf->is_linear) {
       info.va = va + surf_offset + surf->u.gfx9.offset[subresource.mipLevel];
@@ -312,26 +309,79 @@ radv_sdma_get_surf(const struct radv_device *const device, const struct radv_ima
       info.info_dword = radv_sdma_get_tiled_info_dword(device, image, surf, subresource);
       info.header_dword = radv_sdma_get_tiled_header_dword(device, image, subresource);
 
-      if (pdev->info.sdma_supports_compression &&
-          (radv_dcc_enabled(image, subresource.mipLevel) || radv_image_has_htile(image))) {
+      if (pdev->info.gfx_level >= GFX12) {
+         info.is_compressed = binding->bo && binding->bo->gfx12_allow_dcc;
+      } else if (pdev->info.sdma_supports_compression &&
+                 (radv_dcc_enabled(image, subresource.mipLevel) || radv_htile_enabled(image, subresource.mipLevel))) {
+         info.is_compressed = true;
+      }
+
+      if (info.is_compressed) {
          info.meta_va = va + surf->meta_offset;
-         info.meta_config = radv_sdma_get_metadata_config(device, image, surf, subresource, aspect_mask);
+         info.meta_config = radv_sdma_get_metadata_config(device, image, surf, subresource);
       }
    }
 
    return info;
 }
 
-static void
+void
 radv_sdma_emit_nop(const struct radv_device *device, struct radeon_cmdbuf *cs)
 {
    /* SDMA NOP acts as a fence command and causes the SDMA engine to wait for pending copy operations. */
    radeon_check_space(device->ws, cs, 1);
-   radeon_emit(cs, SDMA_PACKET(SDMA_OPCODE_NOP, 0, 0));
+   radeon_begin(cs);
+   radeon_emit(SDMA_PACKET(SDMA_OPCODE_NOP, 0, 0));
+   radeon_end();
 }
 
 void
-radv_sdma_copy_buffer(const struct radv_device *device, struct radeon_cmdbuf *cs, uint64_t src_va, uint64_t dst_va,
+radv_sdma_emit_write_timestamp(struct radeon_cmdbuf *cs, uint64_t va)
+{
+   radeon_begin(cs);
+   radeon_emit(SDMA_PACKET(SDMA_OPCODE_TIMESTAMP, SDMA_TS_SUB_OPCODE_GET_GLOBAL_TIMESTAMP, 0));
+   radeon_emit(va);
+   radeon_emit(va >> 32);
+   radeon_end();
+}
+
+void
+radv_sdma_emit_fence(struct radeon_cmdbuf *cs, uint64_t va, uint32_t fence)
+{
+   radeon_begin(cs);
+   radeon_emit(SDMA_PACKET(SDMA_OPCODE_FENCE, 0, SDMA_FENCE_MTYPE_UC));
+   radeon_emit(va);
+   radeon_emit(va >> 32);
+   radeon_emit(fence);
+   radeon_end();
+}
+
+void
+radv_sdma_emit_wait_mem(struct radeon_cmdbuf *cs, uint32_t op, uint64_t va, uint32_t ref, uint32_t mask)
+{
+   radeon_begin(cs);
+   radeon_emit(SDMA_PACKET(SDMA_OPCODE_POLL_REGMEM, 0, 0) | op << 28 | SDMA_POLL_MEM);
+   radeon_emit(va);
+   radeon_emit(va >> 32);
+   radeon_emit(ref);
+   radeon_emit(mask);
+   radeon_emit(SDMA_POLL_INTERVAL_160_CLK | SDMA_POLL_RETRY_INDEFINITELY << 16);
+   radeon_end();
+}
+
+void
+radv_sdma_emit_write_data_head(struct radeon_cmdbuf *cs, uint64_t va, uint32_t count)
+{
+   radeon_begin(cs);
+   radeon_emit(SDMA_PACKET(SDMA_OPCODE_WRITE, SDMA_WRITE_SUB_OPCODE_LINEAR, 0));
+   radeon_emit(va);
+   radeon_emit(va >> 32);
+   radeon_emit(count - 1);
+   radeon_end();
+}
+
+void
+radv_sdma_copy_memory(const struct radv_device *device, struct radeon_cmdbuf *cs, uint64_t src_va, uint64_t dst_va,
                       uint64_t size)
 {
    if (size == 0)
@@ -360,23 +410,27 @@ radv_sdma_copy_buffer(const struct radv_device *device, struct radeon_cmdbuf *cs
 
    radeon_check_space(device->ws, cs, ncopy * 7);
 
+   radeon_begin(cs);
+
    for (unsigned i = 0; i < ncopy; i++) {
       unsigned csize = size >= 4 ? MIN2(size & align, max_size_per_packet) : size;
-      radeon_emit(cs, SDMA_PACKET(SDMA_OPCODE_COPY, SDMA_COPY_SUB_OPCODE_LINEAR, 0));
-      radeon_emit(cs, ver >= SDMA_4_0 ? csize - 1 : csize);
-      radeon_emit(cs, 0); /* src/dst endian swap */
-      radeon_emit(cs, src_va);
-      radeon_emit(cs, src_va >> 32);
-      radeon_emit(cs, dst_va);
-      radeon_emit(cs, dst_va >> 32);
+      radeon_emit(SDMA_PACKET(SDMA_OPCODE_COPY, SDMA_COPY_SUB_OPCODE_LINEAR, 0));
+      radeon_emit(ver >= SDMA_4_0 ? csize - 1 : csize);
+      radeon_emit(0); /* src/dst endian swap */
+      radeon_emit(src_va);
+      radeon_emit(src_va >> 32);
+      radeon_emit(dst_va);
+      radeon_emit(dst_va >> 32);
       dst_va += csize;
       src_va += csize;
       size -= csize;
    }
+
+   radeon_end();
 }
 
 void
-radv_sdma_fill_buffer(const struct radv_device *device, struct radeon_cmdbuf *cs, const uint64_t va,
+radv_sdma_fill_memory(const struct radv_device *device, struct radeon_cmdbuf *cs, const uint64_t va,
                       const uint64_t size, const uint32_t value)
 {
    const struct radv_physical_device *pdev = radv_device_physical(device);
@@ -395,18 +449,21 @@ radv_sdma_fill_buffer(const struct radv_device *device, struct radeon_cmdbuf *cs
    const unsigned num_packets = DIV_ROUND_UP(size, max_fill_bytes);
    ASSERTED unsigned cdw_max = radeon_check_space(device->ws, cs, num_packets * 5);
 
+   radeon_begin(cs);
+
    for (unsigned i = 0; i < num_packets; ++i) {
       const uint64_t offset = i * max_fill_bytes;
       const uint64_t fill_bytes = MIN2(size - offset, max_fill_bytes);
       const uint64_t fill_va = va + offset;
 
-      radeon_emit(cs, constant_fill_header);
-      radeon_emit(cs, fill_va);
-      radeon_emit(cs, fill_va >> 32);
-      radeon_emit(cs, value);
-      radeon_emit(cs, fill_bytes - 1); /* Must be programmed in bytes, even if the fill is done in dwords. */
+      radeon_emit(constant_fill_header);
+      radeon_emit(fill_va);
+      radeon_emit(fill_va >> 32);
+      radeon_emit(value);
+      radeon_emit(fill_bytes - 1); /* Must be programmed in bytes, even if the fill is done in dwords. */
    }
 
+   radeon_end();
    assert(cs->cdw <= cdw_max);
 }
 
@@ -442,20 +499,22 @@ radv_sdma_emit_copy_linear_sub_window(const struct radv_device *device, struct r
 
    ASSERTED unsigned cdw_end = radeon_check_space(device->ws, cs, 13);
 
-   radeon_emit(cs, SDMA_PACKET(SDMA_OPCODE_COPY, SDMA_COPY_SUB_OPCODE_LINEAR_SUB_WINDOW, 0) | util_logbase2(src->bpp)
-                                                                                                 << 29);
-   radeon_emit(cs, src->va);
-   radeon_emit(cs, src->va >> 32);
-   radeon_emit(cs, src_off.x | src_off.y << 16);
-   radeon_emit(cs, src_off.z | (src_pitch - 1) << (ver >= SDMA_7_0 ? 16 : 13));
-   radeon_emit(cs, src_slice_pitch - 1);
-   radeon_emit(cs, dst->va);
-   radeon_emit(cs, dst->va >> 32);
-   radeon_emit(cs, dst_off.x | dst_off.y << 16);
-   radeon_emit(cs, dst_off.z | (dst_pitch - 1) << (ver >= SDMA_7_0 ? 16 : 13));
-   radeon_emit(cs, dst_slice_pitch - 1);
-   radeon_emit(cs, (ext.width - 1) | (ext.height - 1) << 16);
-   radeon_emit(cs, (ext.depth - 1));
+   radeon_begin(cs);
+   radeon_emit(SDMA_PACKET(SDMA_OPCODE_COPY, SDMA_COPY_SUB_OPCODE_LINEAR_SUB_WINDOW, 0) | util_logbase2(src->bpp)
+                                                                                             << 29);
+   radeon_emit(src->va);
+   radeon_emit(src->va >> 32);
+   radeon_emit(src_off.x | src_off.y << 16);
+   radeon_emit(src_off.z | (src_pitch - 1) << (ver >= SDMA_7_0 ? 16 : 13));
+   radeon_emit(src_slice_pitch - 1);
+   radeon_emit(dst->va);
+   radeon_emit(dst->va >> 32);
+   radeon_emit(dst_off.x | dst_off.y << 16);
+   radeon_emit(dst_off.z | (dst_pitch - 1) << (ver >= SDMA_7_0 ? 16 : 13));
+   radeon_emit(dst_slice_pitch - 1);
+   radeon_emit((ext.width - 1) | (ext.height - 1) << 16);
+   radeon_emit((ext.depth - 1));
+   radeon_end();
 
    assert(cs->cdw == cdw_end);
 }
@@ -469,7 +528,7 @@ radv_sdma_emit_copy_tiled_sub_window(const struct radv_device *device, struct ra
    const struct radv_physical_device *pdev = radv_device_physical(device);
 
    if (!pdev->info.sdma_supports_compression) {
-      assert(!tiled->meta_va);
+      assert(!tiled->is_compressed);
    }
 
    const VkOffset3D linear_off = radv_sdma_pixel_offset_to_blocks(linear->offset, linear->blk_w, linear->blk_h);
@@ -478,7 +537,7 @@ radv_sdma_emit_copy_tiled_sub_window(const struct radv_device *device, struct ra
    const VkExtent3D ext = radv_sdma_pixel_extent_to_blocks(pix_extent, tiled->blk_w, tiled->blk_h);
    const unsigned linear_pitch = radv_sdma_pixels_to_blocks(linear->pitch, tiled->blk_w);
    const unsigned linear_slice_pitch = radv_sdma_pixel_area_to_blocks(linear->slice_pitch, tiled->blk_w, tiled->blk_h);
-   const bool dcc = !!tiled->meta_va;
+   const bool dcc = tiled->is_compressed;
    const bool uses_depth = linear_off.z != 0 || tiled_off.z != 0 || ext.depth != 1;
 
    assert(util_is_power_of_two_nonzero(tiled->bpp));
@@ -486,30 +545,35 @@ radv_sdma_emit_copy_tiled_sub_window(const struct radv_device *device, struct ra
 
    ASSERTED unsigned cdw_end = radeon_check_space(device->ws, cs, 14 + (dcc ? 3 : 0));
 
-   radeon_emit(cs, SDMA_PACKET(SDMA_OPCODE_COPY, SDMA_COPY_SUB_OPCODE_TILED_SUB_WINDOW, 0) | dcc << 19 | detile << 31 |
-                      tiled->header_dword);
-   radeon_emit(cs, tiled->va);
-   radeon_emit(cs, tiled->va >> 32);
-   radeon_emit(cs, tiled_off.x | tiled_off.y << 16);
-   radeon_emit(cs, tiled_off.z | (tiled_ext.width - 1) << 16);
-   radeon_emit(cs, (tiled_ext.height - 1) | (tiled_ext.depth - 1) << 16);
-   radeon_emit(cs, tiled->info_dword);
-   radeon_emit(cs, linear->va);
-   radeon_emit(cs, linear->va >> 32);
-   radeon_emit(cs, linear_off.x | linear_off.y << 16);
-   radeon_emit(cs, linear_off.z | (linear_pitch - 1) << 16);
-   radeon_emit(cs, linear_slice_pitch - 1);
-   radeon_emit(cs, (ext.width - 1) | (ext.height - 1) << 16);
-   radeon_emit(cs, (ext.depth - 1));
+   radeon_begin(cs);
+   radeon_emit(SDMA_PACKET(SDMA_OPCODE_COPY, SDMA_COPY_SUB_OPCODE_TILED_SUB_WINDOW, 0) | dcc << 19 | detile << 31 |
+               tiled->header_dword);
+   radeon_emit(tiled->va);
+   radeon_emit(tiled->va >> 32);
+   radeon_emit(tiled_off.x | tiled_off.y << 16);
+   radeon_emit(tiled_off.z | (tiled_ext.width - 1) << 16);
+   radeon_emit((tiled_ext.height - 1) | (tiled_ext.depth - 1) << 16);
+   radeon_emit(tiled->info_dword);
+   radeon_emit(linear->va);
+   radeon_emit(linear->va >> 32);
+   radeon_emit(linear_off.x | linear_off.y << 16);
+   radeon_emit(linear_off.z | (linear_pitch - 1) << 16);
+   radeon_emit(linear_slice_pitch - 1);
+   radeon_emit((ext.width - 1) | (ext.height - 1) << 16);
+   radeon_emit((ext.depth - 1));
 
-   if (tiled->meta_va) {
-      const unsigned write_compress_enable = !detile;
-      radeon_emit(cs, tiled->meta_va);
-      radeon_emit(cs, tiled->meta_va >> 32);
-      radeon_emit(cs, tiled->meta_config | write_compress_enable << 28);
+   if (tiled->is_compressed) {
+      if (pdev->info.sdma_ip_version >= SDMA_7_0) {
+         radeon_emit(tiled->meta_config | SDMA7_DCC_WRITE_CM(!detile));
+      } else {
+         radeon_emit(tiled->meta_va);
+         radeon_emit(tiled->meta_va >> 32);
+         radeon_emit(tiled->meta_config | SDMA5_DCC_WRITE_COMPRESS(!detile));
+      }
    }
 
-   assert(cs->cdw == cdw_end);
+   radeon_end();
+   assert(cs->cdw <= cdw_end);
 }
 
 static void
@@ -523,21 +587,21 @@ radv_sdma_emit_copy_t2t_sub_window(const struct radv_device *device, struct rade
    assert(pdev->info.sdma_ip_version >= SDMA_4_0);
 
    /* On GFX10+ this supports DCC, but cannot copy a compressed surface to another compressed surface. */
-   assert(!src->meta_va || !dst->meta_va);
+   assert(!src->is_compressed || !dst->is_compressed);
 
    if (pdev->info.sdma_ip_version >= SDMA_4_0 && pdev->info.sdma_ip_version < SDMA_5_0) {
       /* SDMA v4 doesn't support mip_id selection in the T2T copy packet. */
       assert(src->header_dword >> 24 == 0);
       assert(dst->header_dword >> 24 == 0);
       /* SDMA v4 doesn't support any image metadata. */
-      assert(!src->meta_va);
-      assert(!dst->meta_va);
+      assert(!src->is_compressed);
+      assert(!dst->is_compressed);
    }
 
    /* Despite the name, this can indicate DCC or HTILE metadata. */
-   const uint32_t dcc = src->meta_va || dst->meta_va;
+   const uint32_t dcc = src->is_compressed || dst->is_compressed;
    /* 0 = compress (src is uncompressed), 1 = decompress (src is compressed). */
-   const uint32_t dcc_dir = src->meta_va && !dst->meta_va;
+   const uint32_t dcc_dir = src->is_compressed && !dst->is_compressed;
 
    const VkOffset3D src_off = radv_sdma_pixel_offset_to_blocks(src->offset, src->blk_w, src->blk_h);
    const VkOffset3D dst_off = radv_sdma_pixel_offset_to_blocks(dst->offset, dst->blk_w, dst->blk_h);
@@ -550,35 +614,44 @@ radv_sdma_emit_copy_t2t_sub_window(const struct radv_device *device, struct rade
 
    ASSERTED unsigned cdw_end = radeon_check_space(device->ws, cs, 15 + (dcc ? 3 : 0));
 
-   radeon_emit(cs, SDMA_PACKET(SDMA_OPCODE_COPY, SDMA_COPY_SUB_OPCODE_T2T_SUB_WINDOW, 0) | dcc << 19 | dcc_dir << 31 |
-                      src->header_dword);
-   radeon_emit(cs, src->va);
-   radeon_emit(cs, src->va >> 32);
-   radeon_emit(cs, src_off.x | src_off.y << 16);
-   radeon_emit(cs, src_off.z | (src_ext.width - 1) << 16);
-   radeon_emit(cs, (src_ext.height - 1) | (src_ext.depth - 1) << 16);
-   radeon_emit(cs, src->info_dword);
-   radeon_emit(cs, dst->va);
-   radeon_emit(cs, dst->va >> 32);
-   radeon_emit(cs, dst_off.x | dst_off.y << 16);
-   radeon_emit(cs, dst_off.z | (dst_ext.width - 1) << 16);
-   radeon_emit(cs, (dst_ext.height - 1) | (dst_ext.depth - 1) << 16);
-   radeon_emit(cs, dst->info_dword);
-   radeon_emit(cs, (ext.width - 1) | (ext.height - 1) << 16);
-   radeon_emit(cs, (ext.depth - 1));
+   radeon_begin(cs);
+   radeon_emit(SDMA_PACKET(SDMA_OPCODE_COPY, SDMA_COPY_SUB_OPCODE_T2T_SUB_WINDOW, 0) | dcc << 19 | dcc_dir << 31 |
+               src->header_dword);
+   radeon_emit(src->va);
+   radeon_emit(src->va >> 32);
+   radeon_emit(src_off.x | src_off.y << 16);
+   radeon_emit(src_off.z | (src_ext.width - 1) << 16);
+   radeon_emit((src_ext.height - 1) | (src_ext.depth - 1) << 16);
+   radeon_emit(src->info_dword);
+   radeon_emit(dst->va);
+   radeon_emit(dst->va >> 32);
+   radeon_emit(dst_off.x | dst_off.y << 16);
+   radeon_emit(dst_off.z | (dst_ext.width - 1) << 16);
+   radeon_emit((dst_ext.height - 1) | (dst_ext.depth - 1) << 16);
+   radeon_emit(dst->info_dword);
+   radeon_emit((ext.width - 1) | (ext.height - 1) << 16);
+   radeon_emit((ext.depth - 1));
 
-   if (dst->meta_va) {
-      const uint32_t write_compress_enable = 1;
-      radeon_emit(cs, dst->meta_va);
-      radeon_emit(cs, dst->meta_va >> 32);
-      radeon_emit(cs, dst->meta_config | write_compress_enable << 28);
-   } else if (src->meta_va) {
-      radeon_emit(cs, src->meta_va);
-      radeon_emit(cs, src->meta_va >> 32);
-      radeon_emit(cs, src->meta_config);
+   if (pdev->info.sdma_ip_version >= SDMA_7_0) {
+      /* Compress only when dst has DCC. If src has DCC, it automatically decompresses according to
+       * PTE.D (page table bit) even if we don't enable DCC in the packet.
+       */
+      if (dst->is_compressed)
+         radeon_emit(dst->meta_config | SDMA7_DCC_WRITE_CM(1));
+   } else {
+      if (dst->is_compressed) {
+         radeon_emit(dst->meta_va);
+         radeon_emit(dst->meta_va >> 32);
+         radeon_emit(dst->meta_config | SDMA5_DCC_WRITE_COMPRESS(1));
+      } else if (src->is_compressed) {
+         radeon_emit(src->meta_va);
+         radeon_emit(src->meta_va >> 32);
+         radeon_emit(src->meta_config);
+      }
    }
 
-   assert(cs->cdw == cdw_end);
+   radeon_end();
+   assert(cs->cdw <= cdw_end);
 }
 
 void
@@ -664,7 +737,7 @@ radv_sdma_copy_buffer_image_unaligned(const struct radv_device *device, struct r
             const uint64_t buf_va =
                buf->va + slice * buf_slice_pitch_blocks * img.bpp + (row + r) * buf_pitch_blocks * img.bpp;
             const uint64_t tmp_va = tmp.va + r * info.aligned_row_pitch * img.bpp;
-            radv_sdma_copy_buffer(device, cs, to_image ? buf_va : tmp_va, to_image ? tmp_va : buf_va,
+            radv_sdma_copy_memory(device, cs, to_image ? buf_va : tmp_va, to_image ? tmp_va : buf_va,
                                   info.extent_horizontal_blocks * img.bpp);
          }
 
@@ -733,7 +806,7 @@ radv_sdma_use_t2t_scanline_copy(const struct radv_device *device, const struct r
     * It can either compress or decompress, or copy uncompressed images, but it
     * can't copy from a compressed image to another.
     */
-   if (src->meta_va && dst->meta_va)
+   if (src->is_compressed && dst->is_compressed)
       return true;
 
    const bool needs_3d_alignment = src->is_3d && (src->micro_tile_mode == RADEON_MICRO_MODE_DISPLAY ||
