@@ -55,6 +55,7 @@ glsl_type_add_to_function_params(const struct glsl_type *type,
       func->params[(*param_idx)++] = (nir_parameter) {
          .num_components = glsl_get_vector_elements(type),
          .bit_size = glsl_get_bit_size(type),
+         .type = type,
       };
    } else if (glsl_type_is_array_or_matrix(type)) {
       unsigned elems = glsl_get_length(type);
@@ -77,7 +78,10 @@ vtn_ssa_value_add_to_call_params(struct vtn_builder *b,
                                  nir_call_instr *call,
                                  unsigned *param_idx)
 {
-   if (glsl_type_is_vector_or_scalar(value->type)) {
+   if (glsl_type_is_cmat(value->type)) {
+      nir_deref_instr *src_deref = vtn_get_deref_for_ssa_value(b, value);
+      call->params[(*param_idx)++] = nir_src_for_ssa(&src_deref->def);
+   } else if (glsl_type_is_vector_or_scalar(value->type)) {
       call->params[(*param_idx)++] = nir_src_for_ssa(value->def);
    } else {
       unsigned elems = glsl_get_length(value->type);
@@ -108,6 +112,7 @@ function_parameter_decoration_cb(struct vtn_builder *b, struct vtn_value *val,
          case SpvFunctionParameterAttributeNoAlias:
          case SpvFunctionParameterAttributeSext:
          case SpvFunctionParameterAttributeZext:
+         case SpvFunctionParameterAttributeSret:
             break;
 
          case SpvFunctionParameterAttributeByVal:
@@ -129,6 +134,8 @@ function_parameter_decoration_cb(struct vtn_builder *b, struct vtn_value *val,
    case SpvDecorationRelaxedPrecision:
    case SpvDecorationRestrict:
    case SpvDecorationRestrictPointer:
+   case SpvDecorationUniform:
+   case SpvDecorationUniformId:
    case SpvDecorationVolatile:
       break;
 
@@ -146,10 +153,20 @@ vtn_ssa_value_load_function_param(struct vtn_builder *b,
                                   struct vtn_func_arg_info *info,
                                   unsigned *param_idx)
 {
-   if (glsl_type_is_vector_or_scalar(value->type)) {
+   if (glsl_type_is_cmat(value->type)) {
+      nir_variable *copy_var =
+         nir_local_variable_create(b->nb.impl, value->type, "cmat_param_by_value");
+
+      nir_def *param = nir_load_param(&b->nb, (*param_idx)++);
+      nir_deref_instr *copy = nir_build_deref_var(&b->nb, copy_var);
+      nir_cmat_copy(&b->nb, &copy->def, param);
+
+      value->is_variable = true;
+      value->var = copy_var;
+   } else if (glsl_type_is_vector_or_scalar(value->type)) {
       /* if the parameter is passed by value, we need to create a local copy if it's a pointer */
       if (info->by_value && type && type->base_type == vtn_base_type_pointer) {
-         struct vtn_type *pointee_type = type->deref;
+         struct vtn_type *pointee_type = type->pointed;
 
          nir_variable *copy =
             nir_local_variable_create(b->nb.impl, pointee_type->type, NULL);
@@ -239,6 +256,29 @@ function_decoration_cb(struct vtn_builder *b, struct vtn_value *val, int member,
    }
 }
 
+/*
+ * Usually, execution modes are per-shader and handled elsewhere. However, with
+ * create_library we will have modes per-nir_function. We can't represent all
+ * SPIR-V execution modes in nir_function, so this is lossy for multi-entrypoint
+ * SPIR-V. However, we do have workgroup_size in nir_function so we gather that
+ * here. If other execution modes are needed in the multi-entrypoint case, both
+ * nir_function and this callback will need to be extended suitably.
+ */
+static void
+function_execution_mode_cb(struct vtn_builder *b, struct vtn_value *func,
+                           const struct vtn_decoration *mode, void *data)
+{
+   nir_function *nir_func = data;
+
+   if (mode->exec_mode == SpvExecutionModeLocalSize) {
+      vtn_assert(b->shader->info.stage == MESA_SHADER_KERNEL);
+
+      nir_func->workgroup_size[0] = mode->operands[0];
+      nir_func->workgroup_size[1] = mode->operands[1];
+      nir_func->workgroup_size[2] = mode->operands[2];
+   }
+}
+
 bool
 vtn_cfg_handle_prepass_instruction(struct vtn_builder *b, SpvOp opcode,
                                    const uint32_t *w, unsigned count)
@@ -267,6 +307,12 @@ vtn_cfg_handle_prepass_instruction(struct vtn_builder *b, SpvOp opcode,
       nir_function *func =
          nir_function_create(b->shader, ralloc_strdup(b->shader, val->name));
 
+      /* Execution modes are gathered per-function with create_library (here)
+       * but per shader with !create_library (elsewhere).
+       */
+      if (b->options->create_library)
+         vtn_foreach_execution_mode(b, val, function_execution_mode_cb, func);
+
       unsigned num_params = 0;
       for (unsigned i = 0; i < func_type->length; i++)
          num_params += glsl_type_count_function_params(func_type->params[i]->type);
@@ -279,8 +325,19 @@ vtn_cfg_handle_prepass_instruction(struct vtn_builder *b, SpvOp opcode,
       func->dont_inline = b->func->control & SpvFunctionControlDontInlineMask;
       func->is_exported = b->func->linkage == SpvLinkageTypeExport;
 
+      /* This is a bit subtle: if we are compiling a non-library, we will have
+       * exactly one entrypoint. But in library mode, we can have 0, 1, or even
+       * multiple entrypoints. This is OK.
+       *
+       * So, we set is_entrypoint for libraries here (plumbing OpEntryPoint),
+       * but set is_entrypoint elsewhere for graphics shaders.
+       */
+      if (b->options->create_library) {
+         func->is_entrypoint = val->is_entrypoint;
+      }
+
       func->num_params = num_params;
-      func->params = ralloc_array(b->shader, nir_parameter, num_params);
+      func->params = rzalloc_array(b->shader, nir_parameter, num_params);
 
       unsigned idx = 0;
       if (func_type->return_type->base_type != vtn_base_type_void) {
@@ -290,6 +347,8 @@ vtn_cfg_handle_prepass_instruction(struct vtn_builder *b, SpvOp opcode,
          func->params[idx++] = (nir_parameter) {
             .num_components = nir_address_format_num_components(addr_format),
             .bit_size = nir_address_format_bit_size(addr_format),
+            .is_return = true,
+            .type = func_type->return_type->type,
          };
       }
 
@@ -342,6 +401,8 @@ vtn_cfg_handle_prepass_instruction(struct vtn_builder *b, SpvOp opcode,
       struct vtn_ssa_value *ssa = vtn_create_ssa_value(b, type->type);
       struct vtn_value *val = vtn_untyped_value(b, w[2]);
 
+      b->func->nir_func->params[b->func_param_idx].name = val->name;
+
       vtn_foreach_decoration(b, val, function_parameter_decoration_cb, &arg_info);
       vtn_ssa_value_load_function_param(b, ssa, type, &arg_info, &b->func_param_idx);
       vtn_push_ssa_value(b, w[2], ssa);
