@@ -35,6 +35,9 @@ init_shaders(struct vl_compositor *c)
 {
    assert(c);
 
+   if (c->shaders_initialized)
+      return true;
+
    if (c->pipe_cs_composit_supported) {
       if (!vl_compositor_cs_init_shaders(c))
          return false;
@@ -96,6 +99,8 @@ init_shaders(struct vl_compositor *c)
       }
    }
 
+   c->shaders_initialized = true;
+
    return true;
 }
 
@@ -103,6 +108,9 @@ static void cleanup_shaders(struct vl_compositor *c)
 {
    assert(c);
 
+   if (!c->shaders_initialized)
+      return;
+
    if (c->pipe_cs_composit_supported) {
       vl_compositor_cs_cleanup_shaders(c);
    } else if (c->pipe_gfx_supported) {
@@ -147,7 +155,6 @@ init_pipe_state(struct vl_compositor *c)
    sampler.mag_img_filter = PIPE_TEX_FILTER_LINEAR;
    sampler.compare_mode = PIPE_TEX_COMPARE_NONE;
    sampler.compare_func = PIPE_FUNC_ALWAYS;
-
    c->sampler_linear = c->pipe->create_sampler_state(c->pipe, &sampler);
 
    sampler.min_img_filter = PIPE_TEX_FILTER_NEAREST;
@@ -228,8 +235,10 @@ static void cleanup_pipe_state(struct vl_compositor *c)
            c->pipe->delete_blend_state(c->pipe, c->blend_add);
            c->pipe->delete_rasterizer_state(c->pipe, c->rast);
    }
-   c->pipe->delete_sampler_state(c->pipe, c->sampler_linear);
-   c->pipe->delete_sampler_state(c->pipe, c->sampler_nearest);
+   if (c->sampler_linear)
+      c->pipe->delete_sampler_state(c->pipe, c->sampler_linear);
+   if (c->sampler_nearest)
+      c->pipe->delete_sampler_state(c->pipe, c->sampler_nearest);
 }
 
 static bool
@@ -320,21 +329,26 @@ static void
 set_yuv_layer(struct vl_compositor_state *s, struct vl_compositor *c,
               unsigned layer, struct pipe_video_buffer *buffer,
               struct u_rect *src_rect, struct u_rect *dst_rect,
-              bool y, enum vl_compositor_deinterlace deinterlace)
+              enum vl_compositor_plane plane,
+              enum vl_compositor_deinterlace deinterlace)
 {
    struct pipe_sampler_view **sampler_views;
    float half_a_line;
    unsigned i;
+   bool y = plane == VL_COMPOSITOR_PLANE_Y;
 
    assert(s && c && buffer);
 
    assert(layer < VL_COMPOSITOR_MAX_LAYERS);
 
+   if (!init_shaders(c))
+      return;
+
    s->used_layers |= 1 << layer;
    sampler_views = buffer->get_sampler_view_components(buffer);
    for (i = 0; i < 3; ++i) {
       s->layers[layer].samplers[i] = c->sampler_linear;
-      pipe_sampler_view_reference(&s->layers[layer].sampler_views[i], sampler_views[i]);
+      s->layers[layer].sampler_views[i] = sampler_views[i];
    }
 
    calc_src_and_dst(&s->layers[layer], buffer->width, buffer->height,
@@ -366,8 +380,15 @@ set_yuv_layer(struct vl_compositor_state *s, struct vl_compositor *c,
 
    case VL_COMPOSITOR_NONE:
       if (c->pipe_cs_composit_supported) {
-          s->layers[layer].cs = (y) ? c->cs_yuv.progressive.y : c->cs_yuv.progressive.uv;
-          break;
+         if (plane == VL_COMPOSITOR_PLANE_Y)
+            s->layers[layer].cs = c->cs_yuv.progressive.y;
+         else if (plane == VL_COMPOSITOR_PLANE_U)
+            s->layers[layer].cs = c->cs_yuv.progressive.u;
+         else if (plane == VL_COMPOSITOR_PLANE_V)
+            s->layers[layer].cs = c->cs_yuv.progressive.v;
+         else if (plane == VL_COMPOSITOR_PLANE_UV)
+            s->layers[layer].cs = c->cs_yuv.progressive.uv;
+         break;
       }
       FALLTHROUGH;
 
@@ -383,26 +404,38 @@ set_yuv_layer(struct vl_compositor_state *s, struct vl_compositor *c,
 static void
 set_rgb_to_yuv_layer(struct vl_compositor_state *s, struct vl_compositor *c,
                      unsigned layer, struct pipe_sampler_view *v,
-                     struct u_rect *src_rect, struct u_rect *dst_rect, bool y)
+                     struct u_rect *src_rect, struct u_rect *dst_rect,
+                     enum vl_compositor_plane plane)
 {
-   assert(s && c && v);
+   bool y = plane == VL_COMPOSITOR_PLANE_Y;
 
+   assert(s && c && v);
    assert(layer < VL_COMPOSITOR_MAX_LAYERS);
 
+   if (!init_shaders(c))
+      return;
+
    s->used_layers |= 1 << layer;
 
-   if (c->pipe_cs_composit_supported)
-      s->layers[layer].cs = y ? c->cs_rgb_yuv.y : c->cs_rgb_yuv.uv;
-   else if (c->pipe_gfx_supported)
+   if (c->pipe_cs_composit_supported) {
+      if (plane == VL_COMPOSITOR_PLANE_Y)
+         s->layers[layer].cs = c->cs_rgb_yuv.y;
+      else if (plane == VL_COMPOSITOR_PLANE_U)
+         s->layers[layer].cs = c->cs_rgb_yuv.u;
+      else if (plane == VL_COMPOSITOR_PLANE_V)
+         s->layers[layer].cs = c->cs_rgb_yuv.v;
+      else if (plane == VL_COMPOSITOR_PLANE_UV)
+         s->layers[layer].cs = c->cs_rgb_yuv.uv;
+   } else if (c->pipe_gfx_supported)
       s->layers[layer].fs = y ? c->fs_rgb_yuv.y : c->fs_rgb_yuv.uv;
 
    s->layers[layer].samplers[0] = c->sampler_linear;
    s->layers[layer].samplers[1] = NULL;
    s->layers[layer].samplers[2] = NULL;
 
-   pipe_sampler_view_reference(&s->layers[layer].sampler_views[0], v);
-   pipe_sampler_view_reference(&s->layers[layer].sampler_views[1], NULL);
-   pipe_sampler_view_reference(&s->layers[layer].sampler_views[2], NULL);
+   s->layers[layer].sampler_views[0] = v;
+   s->layers[layer].sampler_views[1] = NULL;
+   s->layers[layer].sampler_views[2] = NULL;
 
    calc_src_and_dst(&s->layers[layer], v->texture->width0, v->texture->height0,
                     src_rect ? *src_rect : default_rect(&s->layers[layer]),
@@ -456,9 +489,10 @@ vl_compositor_clear_layers(struct vl_compositor_state *s)
       s->layers[i].viewport.swizzle_z = PIPE_VIEWPORT_SWIZZLE_POSITIVE_Z;
       s->layers[i].viewport.swizzle_w = PIPE_VIEWPORT_SWIZZLE_POSITIVE_W;
       s->layers[i].rotate = VL_COMPOSITOR_ROTATE_0;
+      s->layers[i].mirror = VL_COMPOSITOR_MIRROR_NONE;
 
       for ( j = 0; j < 3; j++)
-         pipe_sampler_view_reference(&s->layers[i].sampler_views[j], NULL);
+         s->layers[i].sampler_views[j] = NULL;
       for ( j = 0; j < 4; ++j)
          s->layers[i].colors[j] = v_one;
    }
@@ -548,11 +582,14 @@ vl_compositor_set_buffer_layer(struct vl_compositor_state *s,
 
    assert(layer < VL_COMPOSITOR_MAX_LAYERS);
 
+   if (!init_shaders(c))
+      return;
+
    s->used_layers |= 1 << layer;
    sampler_views = buffer->get_sampler_view_components(buffer);
    for (i = 0; i < 3; ++i) {
       s->layers[layer].samplers[i] = c->sampler_linear;
-      pipe_sampler_view_reference(&s->layers[layer].sampler_views[i], sampler_views[i]);
+      s->layers[layer].sampler_views[i] = sampler_views[i];
    }
 
    calc_src_and_dst(&s->layers[layer], buffer->width, buffer->height,
@@ -614,6 +651,9 @@ vl_compositor_set_palette_layer(struct vl_compositor_state *s,
 
    assert(layer < VL_COMPOSITOR_MAX_LAYERS);
 
+   if (!init_shaders(c))
+      return;
+
    s->used_layers |= 1 << layer;
 
    s->layers[layer].fs = include_color_conversion ?
@@ -622,9 +662,9 @@ vl_compositor_set_palette_layer(struct vl_compositor_state *s,
    s->layers[layer].samplers[0] = c->sampler_linear;
    s->layers[layer].samplers[1] = c->sampler_nearest;
    s->layers[layer].samplers[2] = NULL;
-   pipe_sampler_view_reference(&s->layers[layer].sampler_views[0], indexes);
-   pipe_sampler_view_reference(&s->layers[layer].sampler_views[1], palette);
-   pipe_sampler_view_reference(&s->layers[layer].sampler_views[2], NULL);
+   s->layers[layer].sampler_views[0] = indexes;
+   s->layers[layer].sampler_views[1] = palette;
+   s->layers[layer].sampler_views[2] = NULL;
    calc_src_and_dst(&s->layers[layer], indexes->texture->width0, indexes->texture->height0,
                     src_rect ? *src_rect : default_rect(&s->layers[layer]),
                     dst_rect ? *dst_rect : default_rect(&s->layers[layer]));
@@ -645,14 +685,20 @@ vl_compositor_set_rgba_layer(struct vl_compositor_state *s,
 
    assert(layer < VL_COMPOSITOR_MAX_LAYERS);
 
+   if (!init_shaders(c))
+      return;
+
    s->used_layers |= 1 << layer;
-   s->layers[layer].fs = c->fs_rgba;
+   if (c->fs_rgba)
+      s->layers[layer].fs = c->fs_rgba;
+   else if (c->cs_rgba)
+      s->layers[layer].cs = c->cs_rgba;
    s->layers[layer].samplers[0] = c->sampler_linear;
    s->layers[layer].samplers[1] = NULL;
    s->layers[layer].samplers[2] = NULL;
-   pipe_sampler_view_reference(&s->layers[layer].sampler_views[0], rgba);
-   pipe_sampler_view_reference(&s->layers[layer].sampler_views[1], NULL);
-   pipe_sampler_view_reference(&s->layers[layer].sampler_views[2], NULL);
+   s->layers[layer].sampler_views[0] = rgba;
+   s->layers[layer].sampler_views[1] = NULL;
+   s->layers[layer].sampler_views[2] = NULL;
    calc_src_and_dst(&s->layers[layer], rgba->texture->width0, rgba->texture->height0,
                     src_rect ? *src_rect : default_rect(&s->layers[layer]),
                     dst_rect ? *dst_rect : default_rect(&s->layers[layer]));
@@ -672,6 +718,16 @@ vl_compositor_set_layer_rotation(struct vl_compositor_state *s,
    s->layers[layer].rotate = rotate;
 }
 
+void
+vl_compositor_set_layer_mirror(struct vl_compositor_state *s,
+                               unsigned layer,
+                               enum vl_compositor_mirror mirror)
+{
+   assert(s);
+   assert(layer < VL_COMPOSITOR_MAX_LAYERS);
+   s->layers[layer].mirror = mirror;
+}
+
 void
 vl_compositor_yuv_deint_full(struct vl_compositor_state *s,
                              struct vl_compositor *c,
@@ -684,24 +740,43 @@ vl_compositor_yuv_deint_full(struct vl_compositor_state *s,
    struct pipe_surface **dst_surfaces;
 
    dst_surfaces = dst->get_surfaces(dst);
-   vl_compositor_clear_layers(s);
 
-   set_yuv_layer(s, c, 0, src, src_rect, NULL, true, deinterlace);
+   set_yuv_layer(s, c, 0, src, src_rect, NULL, VL_COMPOSITOR_PLANE_Y, deinterlace);
    vl_compositor_set_layer_dst_area(s, 0, dst_rect);
    vl_compositor_render(s, c, dst_surfaces[0], NULL, false);
 
-   if (dst_rect) {
-      dst_rect->x0 /= 2;
-      dst_rect->y0 /= 2;
-      dst_rect->x1 /= 2;
-      dst_rect->y1 /= 2;
+   if (dst_surfaces[1]) {
+      bool clear = util_format_get_nr_components(src->buffer_format) == 1;
+      union pipe_color_union clear_color = { .f = {0.5, 0.5} };
+      dst_rect->x0 = util_format_get_plane_width(dst->buffer_format, 1, dst_rect->x0);
+      dst_rect->x1 = util_format_get_plane_width(dst->buffer_format, 1, dst_rect->x1);
+      dst_rect->y0 = util_format_get_plane_height(dst->buffer_format, 1, dst_rect->y0);
+      dst_rect->y1 = util_format_get_plane_height(dst->buffer_format, 1, dst_rect->y1);
+      set_yuv_layer(s, c, 0, src, src_rect, NULL, dst_surfaces[2] ? VL_COMPOSITOR_PLANE_U :
+                    VL_COMPOSITOR_PLANE_UV, deinterlace);
+      vl_compositor_set_layer_dst_area(s, 0, dst_rect);
+      if (clear) {
+         struct u_rect clear_rect = *dst_rect;
+         s->used_layers = 0;
+         vl_compositor_set_clear_color(s, &clear_color);
+         vl_compositor_render(s, c, dst_surfaces[1], &clear_rect, true);
+      } else {
+         vl_compositor_render(s, c, dst_surfaces[1], NULL, false);
+      }
+
+      if (dst_surfaces[2]) {
+         set_yuv_layer(s, c, 0, src, src_rect, NULL, VL_COMPOSITOR_PLANE_V, deinterlace);
+         vl_compositor_set_layer_dst_area(s, 0, dst_rect);
+         if (clear) {
+            struct u_rect clear_rect = *dst_rect;
+            s->used_layers = 0;
+            vl_compositor_set_clear_color(s, &clear_color);
+            vl_compositor_render(s, c, dst_surfaces[2], &clear_rect, true);
+         } else {
+            vl_compositor_render(s, c, dst_surfaces[2], NULL, false);
+         }
+      }
    }
-
-   set_yuv_layer(s, c, 0, src, src_rect, NULL, false, deinterlace);
-   vl_compositor_set_layer_dst_area(s, 0, dst_rect);
-   vl_compositor_render(s, c, dst_surfaces[1], NULL, false);
-
-   s->pipe->flush(s->pipe, NULL, 0);
 }
 
 void
@@ -722,25 +797,28 @@ vl_compositor_convert_rgb_to_yuv(struct vl_compositor_state *s,
    u_sampler_view_default_template(&sv_templ, src_res, src_res->format);
    sv = s->pipe->create_sampler_view(s->pipe, src_res, &sv_templ);
 
-   vl_compositor_clear_layers(s);
-
-   set_rgb_to_yuv_layer(s, c, 0, sv, src_rect, NULL, true);
+   set_rgb_to_yuv_layer(s, c, 0, sv, src_rect, NULL, VL_COMPOSITOR_PLANE_Y);
    vl_compositor_set_layer_dst_area(s, 0, dst_rect);
    vl_compositor_render(s, c, dst_surfaces[0], NULL, false);
 
-   if (dst_rect) {
-      dst_rect->x0 /= 2;
-      dst_rect->y0 /= 2;
-      dst_rect->x1 /= 2;
-      dst_rect->y1 /= 2;
+   if (dst_surfaces[1]) {
+      dst_rect->x0 = util_format_get_plane_width(dst->buffer_format, 1, dst_rect->x0);
+      dst_rect->x1 = util_format_get_plane_width(dst->buffer_format, 1, dst_rect->x1);
+      dst_rect->y0 = util_format_get_plane_height(dst->buffer_format, 1, dst_rect->y0);
+      dst_rect->y1 = util_format_get_plane_height(dst->buffer_format, 1, dst_rect->y1);
+      set_rgb_to_yuv_layer(s, c, 0, sv, src_rect, NULL, dst_surfaces[2] ? VL_COMPOSITOR_PLANE_U :
+                           VL_COMPOSITOR_PLANE_UV);
+      vl_compositor_set_layer_dst_area(s, 0, dst_rect);
+      vl_compositor_render(s, c, dst_surfaces[1], NULL, false);
+
+      if (dst_surfaces[2]) {
+         set_rgb_to_yuv_layer(s, c, 0, sv, src_rect, NULL, VL_COMPOSITOR_PLANE_V);
+         vl_compositor_set_layer_dst_area(s, 0, dst_rect);
+         vl_compositor_render(s, c, dst_surfaces[2], NULL, false);
+      }
    }
 
-   set_rgb_to_yuv_layer(s, c, 0, sv, src_rect, NULL, false);
-   vl_compositor_set_layer_dst_area(s, 0, dst_rect);
-   vl_compositor_render(s, c, dst_surfaces[1], NULL, false);
-   pipe_sampler_view_reference(&sv, NULL);
-
-   s->pipe->flush(s->pipe, NULL, 0);
+   s->pipe->sampler_view_release(s->pipe, sv);
 }
 
 void
@@ -761,17 +839,14 @@ vl_compositor_render(struct vl_compositor_state *s,
 }
 
 bool
-vl_compositor_init(struct vl_compositor *c, struct pipe_context *pipe)
+vl_compositor_init(struct vl_compositor *c, struct pipe_context *pipe, bool compute_only)
 {
    assert(c);
 
    memset(c, 0, sizeof(*c));
 
-   c->pipe_cs_composit_supported = pipe->screen->get_param(pipe->screen, PIPE_CAP_PREFER_COMPUTE_FOR_MULTIMEDIA) &&
-            pipe->screen->get_param(pipe->screen, PIPE_CAP_TGSI_TEX_TXF_LZ) &&
-            pipe->screen->get_param(pipe->screen, PIPE_CAP_TGSI_DIV);
-
-   c->pipe_gfx_supported = pipe->screen->get_param(pipe->screen, PIPE_CAP_GRAPHICS);
+   c->pipe_cs_composit_supported = compute_only || pipe->screen->caps.prefer_compute_for_multimedia;
+   c->pipe_gfx_supported = !compute_only && pipe->screen->caps.graphics;
    c->pipe = pipe;
 
    c->deinterlace = VL_COMPOSITOR_NONE;
@@ -780,11 +855,6 @@ vl_compositor_init(struct vl_compositor *c, struct pipe_context *pipe)
       return false;
    }
 
-   if (!init_shaders(c)) {
-      cleanup_pipe_state(c);
-      return false;
-   }
-
    if (!init_buffers(c)) {
       cleanup_shaders(c);
       cleanup_pipe_state(c);
@@ -818,7 +888,7 @@ vl_compositor_init_state(struct vl_compositor_state *s, struct pipe_context *pip
       pipe->screen,
       PIPE_BIND_CONSTANT_BUFFER,
       PIPE_USAGE_DEFAULT,
-      sizeof(csc_matrix) + 16*sizeof(float) + 2*sizeof(int)
+      sizeof(csc_matrix) + 32 * sizeof(float) + 2 * sizeof(int)
    );
 
    if (!s->shader_params)
