@@ -1,24 +1,6 @@
 /*
- * Copyright (C) 2012 Rob Clark <robclark@freedesktop.org>
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
- * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
- * SOFTWARE.
+ * Copyright © 2012 Rob Clark <robclark@freedesktop.org>
+ * SPDX-License-Identifier: MIT
  *
  * Authors:
  *    Rob Clark <robclark@freedesktop.org>
@@ -671,7 +653,8 @@ fd_stream_output_target_destroy(struct pipe_context *pctx,
 static void
 fd_set_stream_output_targets(struct pipe_context *pctx, unsigned num_targets,
                              struct pipe_stream_output_target **targets,
-                             const unsigned *offsets) in_dt
+                             const unsigned *offsets,
+                             enum mesa_prim output_prim) in_dt
 {
    struct fd_context *ctx = fd_context(pctx);
    struct fd_streamout_stateobj *so = &ctx->streamout;
@@ -734,38 +717,6 @@ fd_bind_compute_state(struct pipe_context *pctx, void *state) in_dt
    fd_context_dirty_shader(ctx, PIPE_SHADER_COMPUTE, FD_DIRTY_SHADER_PROG);
 }
 
-/* TODO pipe_context::set_compute_resources() should DIAF and clover
- * should be updated to use pipe_context::set_constant_buffer() and
- * pipe_context::set_shader_images().  Until then just directly frob
- * the UBO/image state to avoid the rest of the driver needing to
- * know about this bastard api..
- */
-static void
-fd_set_compute_resources(struct pipe_context *pctx, unsigned start,
-                         unsigned count, struct pipe_surface **prscs) in_dt
-{
-   struct fd_context *ctx = fd_context(pctx);
-   struct fd_constbuf_stateobj *so = &ctx->constbuf[PIPE_SHADER_COMPUTE];
-
-   for (unsigned i = 0; i < count; i++) {
-      const uint32_t index = i + start + 1;   /* UBOs start at index 1 */
-
-      if (!prscs) {
-         util_copy_constant_buffer(&so->cb[index], NULL, false);
-         so->enabled_mask &= ~(1 << index);
-      } else if (prscs[i]->format == PIPE_FORMAT_NONE) {
-         struct pipe_constant_buffer cb = {
-               .buffer = prscs[i]->texture,
-         };
-         util_copy_constant_buffer(&so->cb[index], &cb, false);
-         so->enabled_mask |= (1 << index);
-      } else {
-         // TODO images
-         unreachable("finishme");
-      }
-   }
-}
-
 /* used by clover to bind global objects, returning the bo address
  * via handles[n]
  */
@@ -850,7 +801,6 @@ fd_state_init(struct pipe_context *pctx)
 
    if (has_compute(fd_screen(pctx->screen))) {
       pctx->bind_compute_state = fd_bind_compute_state;
-      pctx->set_compute_resources = fd_set_compute_resources;
       pctx->set_global_binding = fd_set_global_binding;
    }
 
