@@ -168,7 +168,7 @@ lp_blit(struct pipe_context *pipe,
    util_blitter_save_vertex_shader(lp->blitter, (void*)lp->vs);
    util_blitter_save_geometry_shader(lp->blitter, (void*)lp->gs);
    util_blitter_save_so_targets(lp->blitter, lp->num_so_targets,
-                     (struct pipe_stream_output_target**)lp->so_targets);
+                     (struct pipe_stream_output_target**)lp->so_targets, MESA_PRIM_UNKNOWN);
    util_blitter_save_rasterizer(lp->blitter, (void*)lp->rasterizer);
    util_blitter_save_viewport(lp->blitter, &lp->viewports[0]);
    util_blitter_save_scissor(lp->blitter, &lp->scissors[0]);
@@ -226,18 +226,10 @@ llvmpipe_create_surface(struct pipe_context *pipe,
       if (llvmpipe_resource_is_texture(pt)) {
          assert(surf_tmpl->u.tex.level <= pt->last_level);
          assert(surf_tmpl->u.tex.first_layer <= surf_tmpl->u.tex.last_layer);
-         ps->width = u_minify(pt->width0, surf_tmpl->u.tex.level);
-         ps->height = u_minify(pt->height0, surf_tmpl->u.tex.level);
          ps->u.tex.level = surf_tmpl->u.tex.level;
          ps->u.tex.first_layer = surf_tmpl->u.tex.first_layer;
          ps->u.tex.last_layer = surf_tmpl->u.tex.last_layer;
       } else {
-         /* setting width as number of elements should get us correct
-          * renderbuffer width
-          */
-         ps->width = surf_tmpl->u.buf.last_element
-                   - surf_tmpl->u.buf.first_element + 1;
-         ps->height = pt->height0;
          ps->u.buf.first_element = surf_tmpl->u.buf.first_element;
          ps->u.buf.last_element = surf_tmpl->u.buf.last_element;
          assert(ps->u.buf.first_element <= ps->u.buf.last_element);
