@@ -45,7 +45,7 @@ extern "C" {
 #define PIPE_H264_MAX_DPB_SIZE        17
 #define PIPE_H265_MAX_NUM_LIST_REF    15
 #define PIPE_H265_MAX_DPB_SIZE        16
-#define PIPE_H265_MAX_SLICES          128
+#define PIPE_H265_MAX_SLICES          600
 #define PIPE_H264_MAX_REFERENCES      16
 #define PIPE_H265_MAX_REFERENCES      15
 #define PIPE_AV1_MAX_REFERENCES       8
@@ -54,7 +54,6 @@ extern "C" {
 #define PIPE_DEFAULT_INTRA_IDR_PERIOD 30
 #define PIPE_H2645_EXTENDED_SAR       255
 #define PIPE_ENC_ROI_REGION_NUM_MAX   32
-#define PIPE_DEFAULT_DECODER_FEEDBACK_TIMEOUT_NS 1000000000
 #define PIPE_H2645_LIST_REF_INVALID_ENTRY 0xff
 #define PIPE_H265_MAX_LONG_TERM_REF_PICS_SPS 32
 #define PIPE_H265_MAX_LONG_TERM_PICS 16
@@ -197,8 +196,7 @@ enum pipe_av1_enc_frame_type
    PIPE_AV1_ENC_FRAME_TYPE_KEY = 0x00,
    PIPE_AV1_ENC_FRAME_TYPE_INTER = 0x01,
    PIPE_AV1_ENC_FRAME_TYPE_INTRA_ONLY = 0x02,
-   PIPE_AV1_ENC_FRAME_TYPE_SWITCH = 0x03,
-   PIPE_AV1_ENC_FRAME_TYPE_SHOW_EXISTING = 0x04
+   PIPE_AV1_ENC_FRAME_TYPE_SWITCH = 0x03
 };
 
 enum pipe_h2645_enc_rate_control_method
@@ -763,6 +761,7 @@ struct pipe_h264_enc_dpb_entry
    bool is_ltr;
    struct pipe_video_buffer *buffer;
    bool evict;
+   enum pipe_h2645_enc_picture_type picture_type;
 };
 
 struct pipe_h264_enc_picture_desc
@@ -1382,8 +1381,8 @@ struct pipe_av1_enc_picture_desc
    uint16_t frame_width_sb;
    uint16_t frame_height_sb;
    uint16_t upscaled_width;
-   uint16_t render_width;
-   uint16_t render_height;
+   uint16_t render_width_minus_1;
+   uint16_t render_height_minus_1;
    uint32_t interpolation_filter;
    uint8_t tx_mode;
    uint8_t compound_reference_mode;
@@ -1626,6 +1625,7 @@ struct pipe_mjpeg_picture_desc
       uint16_t crop_y;
       uint16_t crop_width;
       uint16_t crop_height;
+      uint32_t sampling_factor;
    } picture_parameter;
 
    struct
@@ -1792,6 +1792,8 @@ struct pipe_av1_picture_desc
          uint32_t mono_chrome:1;
          uint32_t ref_frame_mvs:1;
          uint32_t film_grain_params_present:1;
+         uint32_t subsampling_x:1;
+         uint32_t subsampling_y:1;
       } seq_info_fields;
 
       uint32_t current_frame_id;
@@ -1978,6 +1980,16 @@ struct pipe_vpp_desc
    enum pipe_video_vpp_color_standard_type out_colors_standard;
    enum pipe_video_vpp_color_range out_color_range;
    enum pipe_video_vpp_chroma_siting out_chroma_siting;
+
+   enum pipe_video_vpp_color_primaries in_color_primaries;
+   enum pipe_video_vpp_transfer_characteristic in_transfer_characteristics;
+   enum pipe_video_vpp_matrix_coefficients in_matrix_coefficients;
+
+   enum pipe_video_vpp_color_primaries out_color_primaries;
+   enum pipe_video_vpp_transfer_characteristic out_transfer_characteristics;
+   enum pipe_video_vpp_matrix_coefficients out_matrix_coefficients;
+
+   enum pipe_video_vpp_filter_flag filter_flags;
 };
 
 
