@@ -406,13 +406,16 @@ brw_nir_lower_vs_inputs(nir_shader *nir)
     * whether it is a double-precision type or not.
     */
    nir_lower_io(nir, nir_var_shader_in, type_size_vec4,
-                nir_lower_io_lower_64bit_to_32);
+                nir_lower_io_lower_64bit_to_32_new);
 
    /* This pass needs actual constants */
    nir_opt_constant_folding(nir);
 
    nir_io_add_const_offset_to_base(nir, nir_var_shader_in);
 
+   /* Update shader_info::dual_slot_inputs */
+   nir_shader_gather_info(nir, nir_shader_get_entrypoint(nir));
+
    /* The last step is to remap VERT_ATTRIB_* to actual registers */
 
    /* Whether or not we have any system generated values.  gl_DrawID is not
@@ -424,7 +427,14 @@ brw_nir_lower_vs_inputs(nir_shader *nir)
       BITSET_TEST(nir->info.system_values_read, SYSTEM_VALUE_VERTEX_ID_ZERO_BASE) ||
       BITSET_TEST(nir->info.system_values_read, SYSTEM_VALUE_INSTANCE_ID);
 
-   const unsigned num_inputs = util_bitcount64(nir->info.inputs_read);
+   const unsigned num_inputs = util_bitcount64(nir->info.inputs_read) +
+      util_bitcount64(nir->info.inputs_read & nir->info.dual_slot_inputs);
+
+   /* In the following loop, the intrinsic base value is the offset in
+    * register slots (2 slots can make up in single input for double/64bit
+    * values). The io_semantics location field is the offset in terms of
+    * attributes.
+    */
 
    nir_foreach_function_impl(impl, nir) {
       nir_builder b = nir_builder_create(impl);
@@ -453,7 +463,8 @@ brw_nir_lower_vs_inputs(nir_shader *nir)
                   nir_intrinsic_instr_create(nir, nir_intrinsic_load_input);
                load->src[0] = nir_src_for_ssa(nir_imm_int(&b, 0));
 
-               nir_intrinsic_set_base(load, num_inputs);
+               unsigned input_offset = 0;
+               unsigned location = BRW_SVGS_VE_INDEX;
                switch (intrin->intrinsic) {
                case nir_intrinsic_load_first_vertex:
                   nir_intrinsic_set_component(load, 0);
@@ -472,7 +483,8 @@ brw_nir_lower_vs_inputs(nir_shader *nir)
                   /* gl_DrawID and IsIndexedDraw are stored right after
                    * gl_VertexID and friends if any of them exist.
                    */
-                  nir_intrinsic_set_base(load, num_inputs + has_sgvs);
+                  input_offset += has_sgvs ? 1 : 0;
+                  location = BRW_DRAWID_VE_INDEX;
                   if (intrin->intrinsic == nir_intrinsic_load_draw_id)
                      nir_intrinsic_set_component(load, 0);
                   else
@@ -482,6 +494,16 @@ brw_nir_lower_vs_inputs(nir_shader *nir)
                   unreachable("Invalid system value intrinsic");
                }
 
+               /* Position the value behind the app's inputs, for base we
+                * account for the double inputs, for the io_semantics
+                * location, it's just the input count.
+                */
+               nir_intrinsic_set_base(load, num_inputs + input_offset);
+               struct nir_io_semantics io = {
+                  .location = VERT_ATTRIB_GENERIC0 + location,
+                  .num_slots = 1,
+               };
+               nir_intrinsic_set_io_semantics(load, io);
                load->num_components = 1;
                nir_def_init(&load->instr, &load->def, 1, 32);
                nir_builder_instr_insert(&b, &load->instr);
@@ -496,9 +518,14 @@ brw_nir_lower_vs_inputs(nir_shader *nir)
                 * number for an attribute by masking out the enabled attributes
                 * before it and counting the bits.
                 */
-               int attr = nir_intrinsic_base(intrin);
-               int slot = util_bitcount64(nir->info.inputs_read &
-                                          BITFIELD64_MASK(attr));
+               const struct nir_io_semantics io =
+                  nir_intrinsic_io_semantics(intrin);
+               const int attr = nir_intrinsic_base(intrin);
+               const int slot = util_bitcount64(nir->info.inputs_read &
+                                                BITFIELD64_MASK(attr)) +
+                                util_bitcount64(nir->info.dual_slot_inputs &
+                                                BITFIELD64_MASK(attr)) +
+                                io.high_dvec2;
                nir_intrinsic_set_base(intrin, slot);
                break;
             }
@@ -850,8 +877,15 @@ brw_nir_optimize(nir_shader *nir,
        * indices will nearly always be in bounds and the cost of the load is
        * low.  Therefore there shouldn't be a performance benefit to avoid it.
        */
-      LOOP_OPT(nir_opt_peephole_select, 0, true, false);
-      LOOP_OPT(nir_opt_peephole_select, 8, true, true);
+      nir_opt_peephole_select_options peephole_select_options = {
+         .limit = 0,
+         .indirect_load_ok = true,
+      };
+      LOOP_OPT(nir_opt_peephole_select, &peephole_select_options);
+
+      peephole_select_options.limit = 8;
+      peephole_select_options.expensive_alu_ok = true;
+      LOOP_OPT(nir_opt_peephole_select, &peephole_select_options);
 
       LOOP_OPT(nir_opt_intrinsics);
       LOOP_OPT(nir_opt_idiv_const, 32);
@@ -886,7 +920,12 @@ brw_nir_optimize(nir_shader *nir,
          LOOP_OPT(nir_opt_dce);
       }
       LOOP_OPT_NOT_IDEMPOTENT(nir_opt_if, nir_opt_if_optimize_phi_true_false);
-      LOOP_OPT(nir_opt_conditional_discard);
+
+      nir_opt_peephole_select_options peephole_discard_options = {
+         .limit = 0,
+         .discard_ok = true,
+      };
+      LOOP_OPT(nir_opt_peephole_select, &peephole_discard_options);
       if (nir->options->max_unroll_iterations != 0) {
          LOOP_OPT_NOT_IDEMPOTENT(nir_opt_loop_unroll);
       }
@@ -952,8 +991,7 @@ lower_bit_size_callback(const nir_instr *instr, UNUSED void *data)
       case nir_op_fcos:
          return 0;
       case nir_op_isign:
-         assert(!"Should have been lowered by nir_opt_algebraic.");
-         return 0;
+         unreachable("Should have been lowered by nir_opt_algebraic.");
       default:
          if (nir_op_infos[alu->op].num_inputs >= 2 &&
              alu->def.bit_size == 8)
@@ -1045,6 +1083,15 @@ lower_xehp_tg4_offset_filter(const nir_instr *instr, UNUSED const void *data)
    if (offset_index < 0)
       return false;
 
+   /* When we have LOD & offset, we can pack both (see
+    * intel_nir_lower_texture.c pack_lod_or_bias_and_offset)
+    */
+   bool has_lod =
+      nir_tex_instr_src_index(tex, nir_tex_src_lod) != -1 ||
+      nir_tex_instr_src_index(tex, nir_tex_src_bias) != -1;
+   if (has_lod)
+      return false;
+
    if (!nir_src_is_const(tex->src[offset_index].src))
       return true;
 
@@ -1090,39 +1137,6 @@ brw_preprocess_nir(const struct brw_compiler *compiler, nir_shader *nir,
    if (intel_needs_workaround(devinfo, 1806565034) && !opts->robust_image_access)
       OPT(intel_nir_clamp_image_1d_2d_array_sizes);
 
-   const struct intel_nir_lower_texture_opts intel_tex_options = {
-      .combined_lod_or_bias_and_offset = compiler->devinfo->ver >= 20,
-   };
-   OPT(intel_nir_lower_texture, &intel_tex_options);
-
-   const nir_lower_tex_options tex_options = {
-      .lower_txp = ~0,
-      .lower_txf_offset = true,
-      .lower_rect_offset = true,
-      .lower_txd_cube_map = true,
-      /* For below, See bspec 45942, "Enable new message layout for cube array" */
-      .lower_txd_3d = devinfo->verx10 >= 125,
-      .lower_txd_array = devinfo->verx10 >= 125,
-      .lower_txb_shadow_clamp = true,
-      .lower_txd_shadow_clamp = true,
-      .lower_txd_offset_clamp = true,
-      .lower_tg4_offsets = true,
-      .lower_txs_lod = true, /* Wa_14012320009 */
-      .lower_offset_filter =
-         devinfo->verx10 >= 125 ? lower_xehp_tg4_offset_filter : NULL,
-      .lower_invalid_implicit_lod = true,
-   };
-
-   /* In the case where TG4 coords are lowered to offsets and we have a
-    * lower_xehp_tg4_offset_filter lowering those offsets further, we need to
-    * rerun the pass because the instructions inserted by the first lowering
-    * are not visible during that first pass.
-    */
-   if (OPT(nir_lower_tex, &tex_options)) {
-      OPT(intel_nir_lower_texture, &intel_tex_options);
-      OPT(nir_lower_tex, &tex_options);
-   }
-
    OPT(nir_normalize_cubemap_coords);
 
    OPT(nir_lower_global_vars_to_local);
@@ -1393,17 +1407,6 @@ brw_nir_link_shaders(const struct brw_compiler *compiler,
       NIR_PASS(_, producer, nir_lower_global_vars_to_local);
       NIR_PASS(_, consumer, nir_lower_global_vars_to_local);
 
-      /* The backend might not be able to handle indirects on
-       * temporaries so we need to lower indirects on any of the
-       * varyings we have demoted here.
-       */
-      NIR_PASS(_, producer, nir_lower_indirect_derefs,
-                  brw_nir_no_indirect_mask(compiler, producer->info.stage),
-                  UINT32_MAX);
-      NIR_PASS(_, consumer, nir_lower_indirect_derefs,
-                  brw_nir_no_indirect_mask(compiler, consumer->info.stage),
-                  UINT32_MAX);
-
       brw_nir_optimize(producer, devinfo);
       brw_nir_optimize(consumer, devinfo);
 
@@ -1555,6 +1558,9 @@ get_mem_access_size_align(nir_intrinsic_op intrin, uint8_t bytes,
                           const void *cb_data)
 {
    const uint32_t align = nir_combined_align(align_mul, align_offset);
+   const struct brw_mem_access_cb_data *mem_cb_data =
+      (struct brw_mem_access_cb_data *)cb_data;
+   const struct intel_device_info *devinfo = mem_cb_data->devinfo;
 
    switch (intrin) {
    case nir_intrinsic_load_ssbo:
@@ -1622,13 +1628,27 @@ get_mem_access_size_align(nir_intrinsic_op intrin, uint8_t bytes,
       };
    } else {
       bytes = MIN2(bytes, 16);
-      return (nir_mem_access_size_align) {
-         .bit_size = 32,
-         .num_components = is_scratch ? 1 :
-                           is_load ? DIV_ROUND_UP(bytes, 4) : bytes / 4,
-         .align = 4,
-         .shift = nir_mem_access_shift_method_scalar,
-      };
+
+      /* With UGM LSC dataport, we don't need to lower 64bit data access into
+       * two 32bit single vector access since it supports direct 64bit data
+       * operation.
+       */
+      if (devinfo->has_lsc && align == 8 && bit_size == 64) {
+         return (nir_mem_access_size_align) {
+            .bit_size = bit_size,
+            .num_components = bytes / 8,
+            .align = bit_size / 8,
+            .shift = nir_mem_access_shift_method_scalar,
+         };
+      } else {
+         return (nir_mem_access_size_align) {
+            .bit_size = 32,
+            .num_components = is_scratch ? 1 :
+                              is_load ? DIV_ROUND_UP(bytes, 4) : bytes / 4,
+            .align = 4,
+            .shift = nir_mem_access_shift_method_scalar,
+         };
+      }
    }
 }
 
@@ -1664,7 +1684,6 @@ brw_vectorize_lower_mem_access(nir_shader *nir,
     *   - fewer send messages
     *   - reduced register pressure
     */
-   nir_divergence_analysis(nir);
    if (OPT(intel_nir_blockify_uniform_loads, compiler->devinfo)) {
       OPT(nir_opt_load_store_vectorize, &options);
 
@@ -1685,6 +1704,11 @@ brw_vectorize_lower_mem_access(nir_shader *nir,
       }
    }
 
+
+   struct brw_mem_access_cb_data cb_data = {
+      .devinfo = compiler->devinfo,
+   };
+
    nir_lower_mem_access_bit_sizes_options mem_access_options = {
       .modes = nir_var_mem_ssbo |
                nir_var_mem_constant |
@@ -1694,6 +1718,7 @@ brw_vectorize_lower_mem_access(nir_shader *nir,
                nir_var_mem_global |
                nir_var_mem_shared,
       .callback = get_mem_access_size_align,
+      .cb_data = &cb_data,
    };
    OPT(nir_lower_mem_access_bit_sizes, &mem_access_options);
 
@@ -1736,6 +1761,37 @@ brw_postprocess_nir(nir_shader *nir, const struct brw_compiler *compiler,
 
    UNUSED bool progress; /* Written by OPT */
 
+   const nir_lower_tex_options tex_options = {
+      .lower_txp = ~0,
+      .lower_txf_offset = true,
+      .lower_rect_offset = true,
+      .lower_txd_cube_map = true,
+      /* For below, See bspec 45942, "Enable new message layout for cube array" */
+      .lower_txd_3d = devinfo->verx10 >= 125,
+      .lower_txd_array = devinfo->verx10 >= 125,
+      .lower_txb_shadow_clamp = true,
+      .lower_txd_shadow_clamp = true,
+      .lower_txd_offset_clamp = true,
+      .lower_tg4_offsets = true,
+      .lower_txs_lod = true, /* Wa_14012320009 */
+      .lower_offset_filter =
+         devinfo->verx10 >= 125 ? lower_xehp_tg4_offset_filter : NULL,
+      .lower_txd_clamp_bindless_sampler = true,
+      .lower_txd_clamp_if_sampler_index_not_lt_16 = true,
+      .lower_invalid_implicit_lod = true,
+      .lower_index_to_offset = true,
+   };
+
+   /* In the case where TG4 coords are lowered to offsets and we have a
+    * lower_xehp_tg4_offset_filter lowering those offsets further, we need to
+    * rerun the pass because the instructions inserted by the first lowering
+    * are not visible during that first pass.
+    */
+   if (OPT(nir_lower_tex, &tex_options))
+      OPT(nir_lower_tex, &tex_options);
+
+   OPT(brw_nir_lower_texture, devinfo);
+
    OPT(intel_nir_lower_sparse_intrinsics);
 
    OPT(nir_lower_bit_size, lower_bit_size_callback, (void *)compiler);
@@ -1756,6 +1812,9 @@ brw_postprocess_nir(nir_shader *nir, const struct brw_compiler *compiler,
       OPT(nir_lower_idiv, &options);
    }
 
+   if (devinfo->ver >= 30)
+      NIR_PASS(_, nir, brw_nir_lower_sample_index_in_coord);
+
    if (gl_shader_stage_can_set_fragment_shading_rate(nir->info.stage))
       NIR_PASS(_, nir, intel_nir_lower_shading_rate_output);
 
@@ -1813,8 +1872,14 @@ brw_postprocess_nir(nir_shader *nir, const struct brw_compiler *compiler,
        * instruction from one of the branches of the if-statement, so now it
        * might be under the threshold of conversion to bcsel.
        */
-      OPT(nir_opt_peephole_select, 0, false, false);
-      OPT(nir_opt_peephole_select, 1, false, true);
+      nir_opt_peephole_select_options peephole_select_options = {
+         .limit = 0,
+      };
+      OPT(nir_opt_peephole_select, &peephole_select_options);
+
+      peephole_select_options.limit = 1;
+      peephole_select_options.expensive_alu_ok = true;
+      OPT(nir_opt_peephole_select, &peephole_select_options);
    }
 
    do {
@@ -1853,9 +1918,6 @@ brw_postprocess_nir(nir_shader *nir, const struct brw_compiler *compiler,
    OPT(nir_opt_move, nir_move_comparisons);
    OPT(nir_opt_dead_cf);
 
-   bool divergence_analysis_dirty = false;
-   NIR_PASS_V(nir, nir_divergence_analysis);
-
    static const nir_lower_subgroups_options subgroups_options = {
       .ballot_bit_size = 32,
       .ballot_components = 1,
@@ -1870,8 +1932,6 @@ brw_postprocess_nir(nir_shader *nir, const struct brw_compiler *compiler,
 
       if (OPT(nir_lower_int64))
          brw_nir_optimize(nir, devinfo);
-
-      divergence_analysis_dirty = true;
    }
 
    /* nir_opt_uniform_subgroup can create some operations (e.g.,
@@ -1894,18 +1954,24 @@ brw_postprocess_nir(nir_shader *nir, const struct brw_compiler *compiler,
       OPT(nir_lower_subgroups, &subgroups_options);
    }
 
-   /* Run intel_nir_lower_conversions only after the last tiem
+   /* Run fsign lowering again after the last time brw_nir_optimize is called.
+    * As is the case with conversion lowering (below), brw_nir_optimize can
+    * create additional fsign instructions.
+    */
+   if (OPT(brw_nir_lower_fsign))
+      OPT(nir_opt_dce);
+
+   /* Run nir_split_conversions only after the last tiem
     * brw_nir_optimize is called. Various optimizations invoked there can
     * rematerialize the conversions that the lowering pass eliminates.
     */
-   OPT(intel_nir_lower_conversions);
+   const nir_split_conversions_options split_conv_opts = {
+      .callback = intel_nir_split_conversions_cb,
+   };
+   OPT(nir_split_conversions, &split_conv_opts);
 
    /* Do this only after the last opt_gcm. GCM will undo this lowering. */
    if (nir->info.stage == MESA_SHADER_FRAGMENT) {
-      if (divergence_analysis_dirty) {
-         NIR_PASS_V(nir, nir_divergence_analysis);
-      }
-
       OPT(intel_nir_lower_non_uniform_barycentric_at_sample);
    }
 
@@ -2028,19 +2094,6 @@ brw_nir_apply_key(nir_shader *nir,
 {
    bool progress = false;
 
-   nir_lower_tex_options nir_tex_opts = {
-      .lower_txd_clamp_bindless_sampler = true,
-      .lower_txd_clamp_if_sampler_index_not_lt_16 = true,
-      .lower_invalid_implicit_lod = true,
-      .lower_index_to_offset = true,
-   };
-   OPT(nir_lower_tex, &nir_tex_opts);
-
-   const struct intel_nir_lower_texture_opts tex_opts = {
-      .combined_lod_and_array_index = compiler->devinfo->ver >= 20,
-   };
-   OPT(intel_nir_lower_texture, &tex_opts);
-
    const nir_lower_subgroups_options subgroups_options = {
       .subgroup_size = get_subgroup_size(&nir->info, max_subgroup_size),
       .ballot_bit_size = 32,
@@ -2135,11 +2188,15 @@ lsc_op_for_nir_intrinsic(const nir_intrinsic_instr *intrin)
 
    case nir_intrinsic_image_load:
    case nir_intrinsic_bindless_image_load:
-      return LSC_OP_LOAD_CMASK;
+      return nir_intrinsic_image_dim(intrin) == GLSL_SAMPLER_DIM_MS ?
+             LSC_OP_LOAD_CMASK_MSRT :
+             LSC_OP_LOAD_CMASK;
 
    case nir_intrinsic_image_store:
    case nir_intrinsic_bindless_image_store:
-      return LSC_OP_STORE_CMASK;
+      return nir_intrinsic_image_dim(intrin) == GLSL_SAMPLER_DIM_MS ?
+             LSC_OP_STORE_CMASK_MSRT :
+             LSC_OP_STORE_CMASK;
 
    default:
       assert(nir_intrinsic_has_atomic_op(intrin));
@@ -2413,8 +2470,7 @@ brw_nir_move_interpolation_to_top(nir_shader *nir)
 
       progress = progress || impl_progress;
 
-      nir_metadata_preserve(impl, impl_progress ? nir_metadata_control_flow
-                                                : nir_metadata_all);
+      nir_progress(impl_progress, impl, nir_metadata_control_flow);
    }
 
    return progress;
@@ -2469,5 +2525,3 @@ brw_nir_lower_simd(nir_shader *nir, unsigned dispatch_width)
    return nir_shader_lower_instructions(nir, filter_simd, lower_simd,
                                  (void *)(uintptr_t)dispatch_width);
 }
-
-
